/*
Bullet Continuous Collision Detection and Physics Library
Copyright (c) 2003-2006 Erwin Coumans  http://continuousphysics.com/Bullet/

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from the use of this software.
Permission is granted to anyone to use this software for any purpose, 
including commercial applications, and to alter it and redistribute it freely, 
subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
*/

// Ray test and identify triangles demo.
// mihail.isakov@gmail.com

#include "btBulletDynamicsCommon.h"
#include "GLDebugDrawer.h"
#include "GL_ShapeDrawer.h"

#include "GlutStuff.h"

#include "HitTriangleDemo.h"
#include "barrel.h"

const int maxProxies = 32766;
const int maxOverlap = 65535;

btRigidBody * hit_body = 0;
btVector3 hit_rel_pos;
btVector3 shot_imp = btVector3(0,0,0);

static void tick_callback(btDynamicsWorld * w, btScalar ts)
{
	if (hit_body)
	{
		if (hit_body->getCollisionFlags() != btCollisionObject::CF_STATIC_OBJECT)
			hit_body->setActivationState(ACTIVE_TAG);
		hit_body->applyImpulse(shot_imp,hit_rel_pos);
		hit_body = NULL;
	}
}

struct  MyClosestRayResultCallback : public btCollisionWorld::ClosestRayResultCallback
{
const btCollisionShape * m_hitTriangleShape;
int					  m_hitTriangleIndex;
int					  m_hitShapePart;


	MyClosestRayResultCallback (const btVector3 & rayFrom,const btVector3 & rayTo)
		: btCollisionWorld::ClosestRayResultCallback(rayFrom, rayTo),
		  m_hitTriangleShape(NULL),
		  m_hitTriangleIndex(0),
		  m_hitShapePart(0)
	{
	}

	virtual ~MyClosestRayResultCallback()
	{
	}

	virtual btScalar addSingleResult(btCollisionWorld::LocalRayResult & rayResult, bool normalInWorldSpace)
	{
		if (rayResult.m_localShapeInfo)
		{
			m_hitTriangleShape = rayResult.m_collisionObject->getCollisionShape();
			m_hitTriangleIndex = rayResult.m_localShapeInfo->m_triangleIndex;
			m_hitShapePart = rayResult.m_localShapeInfo->m_shapePart;
		} else 
		{
			m_hitTriangleShape = NULL;
			m_hitTriangleIndex = 0;
			m_hitShapePart = 0;
		}
	return ClosestRayResultCallback::addSingleResult(rayResult,normalInWorldSpace);
	}
};

HitTriangleDemo::HitTriangleDemo()
{
setCameraDistance(btScalar(1.0));
m_ground_array = 0;
m_ground_vertices = 0;
}

HitTriangleDemo::~HitTriangleDemo()
{
	int i;
	for (i=m_dynamicsWorld->getNumCollisionObjects()-1; i>=0 ;i--)
	{
		btCollisionObject* obj = m_dynamicsWorld->getCollisionObjectArray()[i];
		btRigidBody* body = btRigidBody::upcast(obj);
		if (body && body->getMotionState())
		{
			delete body->getMotionState();
		}
		m_dynamicsWorld->removeCollisionObject( obj );
		delete obj;
	}

	for (int j=0;j<m_collisionShapes.size();j++)
	{
		btCollisionShape* shape = m_collisionShapes[j];
		delete shape;
	}
m_collisionShapes.clear();

	for (int z = 0; z < hit_tracks.size(); z++)
	{
		HitTrack * ht = hit_tracks[z];
		if (!ht) continue;
		delete ht;
		ht = NULL;
	}
hit_tracks.clear();

	delete m_dynamicsWorld;

	//delete solver
	delete m_constraintSolver;

	//delete broadphase
	delete m_overlappingPairCache;

	//delete dispatcher
	delete m_dispatcher;

	delete m_collisionConfiguration;
}

void HitTriangleDemo::initPhysics()
{
	
m_collisionConfiguration = new btDefaultCollisionConfiguration();
m_dispatcher = new btCollisionDispatcher(m_collisionConfiguration);
btVector3 worldMin(-1000,-1000,-1000);
btVector3 worldMax(1000,1000,1000);
m_overlappingPairCache = new btAxisSweep3(worldMin,worldMax);
m_constraintSolver = new btSequentialImpulseConstraintSolver();
m_dynamicsWorld = new btDiscreteDynamicsWorld(m_dispatcher,
											  m_overlappingPairCache,
											  m_constraintSolver,
											  m_collisionConfiguration);
btGImpactCollisionAlgorithm::registerAlgorithm(m_dispatcher);

btTransform tr;
tr.setIdentity();

#define USE_TRIMESH_GROUND
#ifdef USE_TRIMESH_GROUND
btBvhTriangleMeshShape * groundShape = 0;
int i;

const float TRIANGLE_SIZE=5.f;

int vertStride = sizeof(btVector3);
int indexStride = 3*sizeof(int);

const int NUM_VERTS_X = 20;
const int NUM_VERTS_Y = 20;
const int totalVerts = NUM_VERTS_X*NUM_VERTS_Y;

const int totalTriangles = 2*(NUM_VERTS_X-1)*(NUM_VERTS_Y-1);

m_ground_vertices = new btVector3[totalVerts];
int * gIndices =  new int[totalTriangles*3];

	for ( i=0;i<NUM_VERTS_X;i++)
	{
		for (int j=0;j<NUM_VERTS_Y;j++)
		{
			float wl = .2f;
			//height set to zero, but can also use curved landscape, just uncomment out the code
			float height = 0.f;//20.f*sinf(float(i)*wl)*cosf(float(j)*wl);
			m_ground_vertices[i+j*NUM_VERTS_X].setValue(
				(i-NUM_VERTS_X*0.5f)*TRIANGLE_SIZE,
				height,
				(j-NUM_VERTS_Y*0.5f)*TRIANGLE_SIZE);

		}
	}

int index=0;
	for ( i=0;i<NUM_VERTS_X-1;i++)
	{
		for (int j=0;j<NUM_VERTS_Y-1;j++)
		{
			gIndices[index++] = j*NUM_VERTS_X+i;
			gIndices[index++] = j*NUM_VERTS_X+i+1;
			gIndices[index++] = (j+1)*NUM_VERTS_X+i+1;

			gIndices[index++] = j*NUM_VERTS_X+i;
			gIndices[index++] = (j+1)*NUM_VERTS_X+i+1;
			gIndices[index++] = (j+1)*NUM_VERTS_X+i;
		}
	}
	
	m_ground_array = new btTriangleIndexVertexArray(totalTriangles,
		gIndices,
		indexStride,
		totalVerts,(btScalar*) &m_ground_vertices[0].x(),vertStride);

bool useQuantizedAabbCompression = true;
groundShape = new btBvhTriangleMeshShape(m_ground_array,useQuantizedAabbCompression);
tr.setOrigin(btVector3(0,-4.5f,0));
#else
btStaticPlaneShape * groundShape = new btStaticPlaneShape(plane_normal, btScalar(0.0));
btVector3 plane_normal = btVector3(btScalar(0.0), btScalar(1.0), btScalar(0.0));
tr.setOrigin(btVector3(btScalar(0),btScalar(-4.5),btScalar(0)));
#endif
localCreateRigidBody(0,tr,groundShape);
m_collisionShapes.push_back(groundShape);

//
tr.setOrigin(btVector3(0,1.0f,20.0f));
btTriangleIndexVertexArray * mesh_interface = new btTriangleIndexVertexArray(
														faces_size,
														barrel_ia,
														3*sizeof(int),
														vertices_size,
														(btScalar *) &barrel_va[0][0],
														sizeof(btVector3));

btGImpactMeshShape * shape = new btGImpactMeshShape(mesh_interface);
shape->setMargin(btScalar(0.1));
shape->updateBound();
m_collisionShapes.push_back(shape);
btRigidBody * body = localCreateRigidBody(btScalar(20.0),tr,shape);

	if (m_dynamicsWorld) m_dynamicsWorld->setInternalTickCallback(tick_callback, this);

}


void HitTriangleDemo::clientMoveAndDisplay()
{
glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 
float dt = getDeltaTimeMicroseconds() * 0.000001f;
	
	if (m_dynamicsWorld)
	{
		int maxSimSubSteps = 2;

		int numSimSteps = m_dynamicsWorld->stepSimulation(btScalar(dt),maxSimSubSteps);

	}

int size = hit_tracks.size();
		if (size >= 1 && hit_tracks[size-1]->triangle_exists == 1)
		{
			btRigidBody * body = hit_tracks[size-1]->get_rigid_body();
			btTransform t;
			float m[16];
			float * triangle = hit_tracks[size-1]->get_triangle();
			if (body->getMotionState())
			{
				body->getMotionState()->getWorldTransform(t);
				t.getOpenGLMatrix(m);
			}
			else
			{
				t = body->getWorldTransform();
				t.getOpenGLMatrix(m);
			}

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix(); 
		glMultMatrixf(m);
		glColor4f(1.0f,1.0f,0.0f,1.0f);
		glBegin(GL_TRIANGLES);
		glVertex3f(triangle[0],triangle[1],triangle[2]);
		glVertex3f(triangle[3],triangle[4],triangle[5]);
		glVertex3f(triangle[6],triangle[7],triangle[8]);
		glEnd();
		glPopMatrix();
		}


		for(int i=0;i<hit_tracks.size();i++)
		{
			btRigidBody * body = hit_tracks[i]->get_rigid_body();
			if (body == NULL) continue;
			btTransform t;
			float m[16];
			if (body->getMotionState())
			{
				body->getMotionState()->getWorldTransform(t);
				t.getOpenGLMatrix(m);
			}
			else
			{
				t = body->getWorldTransform();
				t.getOpenGLMatrix(m);
			}

			glMatrixMode(GL_MODELVIEW);
			glPushMatrix(); 
			glMultMatrixf(m);
			if (hit_tracks[i]->get_position() != NULL)
			{
				glColor4f(0.0f,1.0f,1.0f,1.0f);
				glPointSize(5.0f);
				glBegin(GL_POINTS);
				glVertex3f(hit_tracks[i]->get_position()[0],
						   hit_tracks[i]->get_position()[1],
						   hit_tracks[i]->get_position()[2]);
				glEnd();
			}
		glPopMatrix();
		}

renderme();

	if (m_dynamicsWorld) m_dynamicsWorld->debugDrawWorld();
	
glFlush();
glutSwapBuffers();

}

void HitTriangleDemo::keyboardCallback(unsigned char key, int x, int y)
{
	switch (key)
	{
	case ' ':
		{
		hit_body = 0;
		hit_rel_pos = btVector3(0,0,0);
			for (int z = 0; z < hit_tracks.size(); z++)
			{
			HitTrack * ht = hit_tracks[z];
				if (!ht) continue;
			delete ht;
			ht = NULL;
			}
		hit_tracks.clear();
		clientResetScene();
		break;
		}
	default:
		DemoApplication::keyboardCallback(key, x, y);
	}
}

void HitTriangleDemo::mouseFunc(int button, int state, int x, int y)
{
	if (state == 0) 
	{
		m_mouseButtons |= 1<<button;
	} else
	{
		m_mouseButtons = 0;
	}

	m_mouseOldX = x;
	m_mouseOldY = y;

	updateModifierKeys();
	if ((m_modifierKeys& BT_ACTIVE_ALT) && (state==0))
	{
		return;
	}

	btVector3 rayTo = getRayTo(x,y);

	switch (button)
	{
	case 0:
		{
			if (state==0)
			{

				if (m_dynamicsWorld)
				{
					
					MyClosestRayResultCallback rayCallback(m_cameraPosition,rayTo);
					m_dynamicsWorld->rayTest(m_cameraPosition,rayTo,rayCallback);

					if (rayCallback.hasHit())
					{
						btVector3 hit = rayCallback.m_hitPointWorld;
						btRigidBody * body = btRigidBody::upcast(rayCallback.m_collisionObject);
						if (body)
						{
							hit_body = body;
							hit_rel_pos = body->getWorldTransform().inverse()(hit);
							shot_imp = rayTo;
							shot_imp.normalize();
							shot_imp *= 20;

							float p[3] = {hit_rel_pos.getX(),
										  hit_rel_pos.getY(),
										  hit_rel_pos.getZ()};
							HitTrack * ht = new HitTrack();
							ht->set_rigid_body(hit_body);
							ht->set_position(p);
							ht->set_distance( (float)(hit - m_cameraPosition).length() );
							//printf("distance=%f\n", ht->get_distance());
							if (rayCallback.m_hitTriangleShape)
							{
								float triangle[9]; 
								bool triangle_ok = process_triangle(hit_body->getCollisionShape(),
																rayCallback.m_hitTriangleIndex,
																triangle);
								if (triangle_ok)
								{
									ht->triangle_exists = 1;
									ht->set_triangle(triangle);
									ht->set_subpart(rayCallback.m_hitShapePart);
								}
							}
							hit_tracks.push_back(ht);
						}
					}
				}

			}
			break;
		};
	case 1:
		{
			if (state==0)
			{
			} else
			{
			}
			break;	
		}
	case 2:
		{
			if (state==0)
			{
			shootBox(rayTo);
			} else
			{
			}
			break;
		}
	default:
		{
		}
	}

}

bool HitTriangleDemo::process_triangle(btCollisionShape * shape,
									 int hitTriangleIndex,
									 float * triangle)
{
btStridingMeshInterface * meshInterface = NULL;

	if (shape->getShapeType() == GIMPACT_SHAPE_PROXYTYPE)
	{
		meshInterface = (static_cast<btGImpactMeshShape*>(shape))->getMeshInterface();
	}
	else if (shape->getShapeType() == TRIANGLE_MESH_SHAPE_PROXYTYPE)
	{
		meshInterface = (static_cast<btBvhTriangleMeshShape*>(shape))->getMeshInterface();
	}
	else
	{
	return false;
	}

	if (!meshInterface) return false;

unsigned char *vertexbase;
int numverts;
PHY_ScalarType type;
int stride;
unsigned char *indexbase;
int indexstride;
int numfaces;
PHY_ScalarType indicestype;

meshInterface->getLockedVertexIndexBase(
	&vertexbase,
	numverts,
	type,
	stride,
	&indexbase,
	indexstride,
	numfaces,
	indicestype,
	0);

unsigned int * gfxbase = (unsigned int*)(indexbase+hitTriangleIndex*indexstride);
const btVector3 & meshScaling = shape->getLocalScaling();
btVector3 triangle_v[3];
	for (int j=2;j>=0;j--)
	{
		int graphicsindex = indicestype==PHY_SHORT?((unsigned short*)gfxbase)[j]:gfxbase[j];

		btScalar * graphicsbase = (btScalar*)(vertexbase+graphicsindex*stride);

//#define TEST_DEFORM
#ifdef TEST_DEFORM
		if (shape->getShapeType() == GIMPACT_SHAPE_PROXYTYPE)
		{
			// model-specific approximations
			if ( graphicsindex >= 0 && graphicsindex <= 31)
			{
			graphicsbase[0] *= btScalar(0.98);
			graphicsbase[1] *= btScalar(0.98);
			graphicsbase[2] *= btScalar(0.98);
			}
			else if (graphicsindex >= 32 && graphicsindex <= 33)
			{
			graphicsbase[0] *= btScalar(0.91);
			graphicsbase[1] *= btScalar(0.91);
			graphicsbase[2] *= btScalar(0.91);
			}
			else
			{
			graphicsbase[0] *= btScalar(0.88);
			graphicsbase[1] *= btScalar(0.88);
			graphicsbase[2] *= btScalar(0.88);
			}
		}

#endif

	triangle_v[j] = btVector3(graphicsbase[0]*meshScaling.getX(),
							  graphicsbase[1]*meshScaling.getY(),
							  graphicsbase[2]*meshScaling.getZ());	

	}

meshInterface->unLockVertexBase(0);

#ifdef TEST_DEFORM
	if (shape->getShapeType() == GIMPACT_SHAPE_PROXYTYPE)
	{
		btGImpactMeshShape * gimp_shape = static_cast<btGImpactMeshShape*>(shape);
		gimp_shape->postUpdate();
	}
#endif

triangle[0] = triangle_v[0].getX();
triangle[1] = triangle_v[0].getY();
triangle[2] = triangle_v[0].getZ();
triangle[3] = triangle_v[1].getX();
triangle[4] = triangle_v[1].getY();
triangle[5] = triangle_v[1].getZ();
triangle[6] = triangle_v[2].getX();
triangle[7] = triangle_v[2].getY();
triangle[8] = triangle_v[2].getZ();

return true;
}


